/* jshint esversion : 6 */

const gulp = require('gulp');
const sourcemaps = require('gulp-sourcemaps');

const $q = require('app_promises');

const uglify = require('gulp-uglify');


const gulpPlugin = {

    sort: o => () => require('gulp-sort')(o),

    concat: o => () => require('gulp-concat')(o),

    strip: o => () => {

        const strip = require('gulp-strip-comments');

        const s = ("object" === typeof o ? o : null);
        return strip(s);
    },

    wrap: o => () => {

        const wrap = require('gulp-wrap');

        const opt = {
            template: o.template,
            data: o.data || {},
            options: o.options || { parse: false }
        };

        const w = [];
        w.push(opt.template);

        if (Array.isArray(opt.data))
            opt.data.forEach(v => w.push(v));
        else
            w.push(opt.data);

        w.push(opt.options);

        return wrap.apply(wrap, w);
    },

    sourcemaps: _o => {

        const o = _o.write;

        const sourceMapsOpt = [];
        if ("string" === o)
            sourceMapsOpt.push(o);
        else {
            const opt = {};
            sourceMapsOpt[0] = o.dest;
            for (let k in o) {
                if (k !== 'dest')
                    opt[k] = o[k];
            }

            sourceMapsOpt[1] = opt;
        }

        return sourcemaps.apply(sourcemaps, sourceMapsOpt);
    }

};



const _mainTaskDep = [];

let _mainTask = 'default';

exports.setMainTask = n => _mainTask = n;





/**
 * @constructor TaskManager
 * @param {String} taskName "name of task"
 * @param {Number} oder "order of execution"
 */
class TaskManager {
    constructor(taskName, order) {
        this.dep = [];
        this.taskName = taskName;

        this.order = (order || (100 + _mainTaskDep.length));

        _mainTaskDep.push({ name: this.taskName, order: this.order });
    }

    /**
     * @param {Object} o
     * @namespace o
     * @property {String|Array} src "files source"
     * @property {String} dest "destination folder"
     * @property {function} sort "function to order files concatanation. It configures the gulp-sort plugin"
     * @property {String} concat "name of concat file. It configures the gulp-concat plugin"
     * @property {Boolean|Object} strip "if true or it's an object it configures the gulp-strip plugin" 
     * @property {Object} wraè "it configure the gulp-wrap plugin"
     */
    set taskDependency(o) {


        if (this.dep.indexOf(o.task) === -1) {
            this.dep.push(o.task);

            if (!o.order)
                o.order = (100 + this.dep.length);

            gulp.task(o.task, (o.dep || []), () => {
                const q = $q.defer();

                const pipes = [];


                ['sort', 'concat', 'strip', 'wrap', 'sourcemaps'].forEach((v, k) => {
                    if (o.hasOwnProperty(v))
                        pipes.splice(k, 0, gulpPlugin[v](o[v]));
                });



                const init = () => (o.sourcemaps ? gulp.src(o.src).pipe(sourcemaps.init(o.sourcemaps.init)) : gulp.src(o.src));


                pipes.reduce((curr, next) => curr.pipe(next()), init())
                    .pipe(gulp.dest(o.dest))
                    .on('end', () => {
                        console.log('end task :: ' + this.taskName + '/' + o.task);
                        q.resolve();
                    });



                return q.promise;
            });
        }
    }

    start() {
        const dep = this.dep.sort((curr, next) => {
            const a = curr.order;
            const b = next.order;
            return (a < b ? -1 : a > b ? 1 : 0);
        });

        return gulp.task(this.taskName, dep, done => {
            console.log('task ' + this.taskName + ' :: completed');
            done();
        });
    }


    get mainTask() {
        return this.start();
    }
}



/**
 * @method 
 * @return {Array}  list of dependencies of the main task
 * @memberof exports
 */
exports.getMainTaskDependency = () => _mainTaskDep;







/**
 * @method
 * @param {Object} o 
 * @namespace o
 * @property {String} o.taskName "name of task" 
 * @property {Array} o.dep "list of dependencies settings"
 * @return {<TaskManager>}
 */
exports.newTaskManager = o => {
    const c = new TaskManager(o.taskName);

    if (o.dep) {
        if (Array.isArray(o.dep))
            o.dep.forEach(v => c.taskDependency = v);
        else
            c.taskDependency = o.dep;
    }

    c.start();

    return c;
};





const _ready = $q.defer();
Object.defineProperty(exports, 'ready', {
    get: () => _ready.promise
});

/**
 * resolve Promise "ready"
 * @method
 * @return {Promise} 
 */
exports.start = function() {
    const dep = _mainTaskDep.sort((curr, next) => {
        const a = curr.order;
        const b = next.order;
        return (a < b ? -1 : a > b ? 1 : 0);
    }).map(e => e.name);


    console.log(dep);

    gulp.task(_mainTask, dep, done => {
        done();
        console.log('* task ' + _mainTask + ' :: completed');
        _ready.resolve();
    });

    process.nextTick(() => gulp.start());

    return this.ready;
};